/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;

/**
 * <p>
 * Listener that can be registered with a {@link ChannelAccessMonitor}.
 * </p>
 * 
 * <p>
 * The listener is notified of monitor events and monitor errors. Monitor events
 * are typically triggered when the value or alarm state of a monitor changes,
 * but the details depend on event mask used for the monitor and the server
 * implementation.
 * </p>
 * 
 * <p>
 * In general, a listener should not run any operations that require an extended
 * amount of time or that even block. Unless explicitly allowed by the
 * implementation, registering such a listener with an object might result in
 * undefined behavior, including a dead lock.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <ValueType>
 *            The exact type of the Channel Access values that are provided by
 *            the monitor events.
 */
public interface ChannelAccessMonitorListener<ValueType extends ChannelAccessGettableValue<?>> {

    /**
     * Handles a monitor error. A monitor error can be caused by a problem when
     * creating the monitor or a problem on the server side.
     * 
     * @param monitor
     *            monitor that triggered the event.
     * @param status
     *            Channel Access status code associated with the error.
     * @param message
     *            text message describing the error in greater detail (might be
     *            <code>null</code>).
     */
    void monitorError(ChannelAccessMonitor<? extends ValueType> monitor,
            ChannelAccessStatus status, String message);

    /**
     * Handles a monitor event. A monitor event is triggered by the server if it
     * encounters an event that is covered by the monitor's event mask. This
     * method is also called with the last monitor event when the listener is
     * registered, unless no monitor event has been received yet.
     * 
     * @param monitor
     *            monitor that triggered the event.
     * @param value
     *            Channel Access value associated with the monitor event.
     */
    void monitorEvent(ChannelAccessMonitor<? extends ValueType> monitor,
            ValueType value);

}
