/*
 * Copyright 2015-2019 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.beacon;

import java.io.File;
import java.io.IOException;
import java.net.BindException;
import java.net.InetSocketAddress;
import java.net.StandardProtocolFamily;
import java.net.StandardSocketOptions;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.DatagramChannel;
import java.nio.channels.SelectionKey;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.apache.commons.lang3.tuple.Pair;
import org.apache.commons.lang3.tuple.Triple;

import com.aquenos.epics.jackie.client.ClientThreadingStrategy;
import com.aquenos.epics.jackie.client.beacon.internal.BeaconServerStatus;
import com.aquenos.epics.jackie.client.resolver.internal.UdpResolver;
import com.aquenos.epics.jackie.common.exception.MalformedMessageException;
import com.aquenos.epics.jackie.common.exception.UnsupportedMessageTypeException;
import com.aquenos.epics.jackie.common.io.ByteBufferByteSource;
import com.aquenos.epics.jackie.common.io.ChannelProcessor;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;
import com.aquenos.epics.jackie.common.io.TimerProcessor;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessBeaconMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessageCodec;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;
import com.aquenos.epics.jackie.common.util.Inet4AddressUtil;
import com.aquenos.epics.jackie.common.util.LinkedWeakIdentityHashSet;

/**
 * <p>
 * Default implementation of a {@link BeaconDetector}.
 * </p>
 * 
 * <p>
 * This class provides an implementation that should be suitable for most use
 * cases, as most important parameters can be configured by through a
 * {@link BeaconDetectorConfiguration}. It creates a UDP socket on which it
 * listens for beacon messages. It tries to connect to an existing Channel
 * Access beacon repeater and, if it cannot find such a repeater on the local
 * host, tries to start it. If it cannot start a repeater, it will start a
 * repeater internally. However, this repeater will only run as long as this
 * beacon detector exists.
 * </p>
 * 
 * <p>
 * In order to work, an object of this class needs to be registered with a
 * {@link CommunicationController}. Typically, code using this class will not
 * instantiate the communication controller directly but instead use a
 * {@link ClientThreadingStrategy} which will create the appropriate controller
 * and pass it to this class when it is registered through the
 * {@link ClientThreadingStrategy#setBeaconDetector(CommunicationProcessor)
 * setBeaconDetector(...)} method.
 * </p>
 * 
 * <p>
 * This implementation tries to stay as close to the implementation in the C
 * library as reasonably possible - as far as the logic for deciding whether a
 * received beacon constitutes a beacon anomaly is concerned.
 * </p>
 * 
 * <p>
 * This beacon detector tries to connect to a running Channel Access repeater.
 * If no repeater is running, it tries to find the <code>caRepeater</code>
 * executable and start it. If the executable cannot be found or the repeater
 * cannot be started, it falls back to creating a {@link Repeater} instance
 * within the Java VM. The lifetime of that instance is not bound to the
 * lifetime of this beacon detector instance. It will keep running until the
 * Java VM is shutdown.
 * </p>
 * 
 * <p>
 * As far as the user-side is concerned, this class is thread safe. This means
 * that the methods provided by the {@link BeaconDetector} interface can be
 * called from any thread, at any time, even before this object has been
 * registered with a communication controller. However, calling those methods
 * will have no effect until the registration with the communication controller
 * is completed.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class DefaultBeaconDetector
        implements CommunicationProcessor, BeaconDetector {

    private static final ByteBuffer EMPTY_BYTE_BUFFER = ByteBuffer.allocate(0);

    private LinkedWeakIdentityHashSet<BeaconAnomalyListener> anomalyListeners = new LinkedWeakIdentityHashSet<BeaconDetector.BeaconAnomalyListener>();
    private ConcurrentLinkedQueue<Pair<BeaconAnomalyListener, Boolean>> anomalyListenerUpdates = new ConcurrentLinkedQueue<Pair<BeaconAnomalyListener, Boolean>>();
    private long beaconDetectorStartTime = 0L;
    private ConcurrentLinkedQueue<Triple<InetSocketAddress, BeaconListener, Boolean>> beaconListenerUpdates = new ConcurrentLinkedQueue<Triple<InetSocketAddress, BeaconListener, Boolean>>();
    private HashMap<InetSocketAddress, BeaconServerStatus> beaconStatusMap = new HashMap<InetSocketAddress, BeaconServerStatus>();
    private DatagramChannel channel;
    private SelectionKey channelSelectionKey;
    private ChannelAccessMessageCodec codec = new ChannelAccessMessageCodec();
    private CommunicationController communicationController;
    private BeaconDetectorConfiguration configuration;
    private long lastBeaconStatusCleanupTime = System.currentTimeMillis();
    private long latestNextMessageTime = 0L;
    private ByteBuffer receiveBuffer = ByteBuffer.allocateDirect(1024);
    private ByteBufferByteSource receiveSource = new ByteBufferByteSource();
    private InetSocketAddress repeaterAddress;
    private TimerProcessor timerProcessor = new TimerProcessor() {
        @Override
        public void processTimer() {
            DefaultBeaconDetector.this.processTimer();
        }
    };

    /**
     * Creates a beacon detector using the specified configuration. The beacon
     * detector is not fully functional until it is registered with a
     * {@link CommunicationController}.
     * 
     * @param configuration
     *            beacon detector configuration (never <code>null</code>).
     */
    public DefaultBeaconDetector(BeaconDetectorConfiguration configuration) {
        this.configuration = configuration;
        this.repeaterAddress = new InetSocketAddress(
                Inet4AddressUtil.LOOPBACK_ADDRESS,
                configuration.getRepeaterPort());
        boolean successful = false;
        try {
            this.channel = DatagramChannel.open(StandardProtocolFamily.INET);
            this.channel.socket().bind(null);
            // Set channel to non-blocking mode. This is critical, because we
            // cannot use selectors with a blocking channel.
            this.channel.configureBlocking(false);
            successful = true;
        } catch (IOException e) {
            throw new RuntimeException(
                    "Could not create the UDP socket for the beacon detector.",
                    e);
        } finally {
            if (!successful) {
                try {
                    if (this.channel != null) {
                        this.channel.close();
                    }
                } catch (IOException e) {
                    // We ignore any exception that might be thrown while
                    // closing the channel.
                }
            }
        }
    }

    private void cleanUpBeaconStatusMap() {
        // Remove old entries from the beacon status map if more than half an
        // hour as passed since the last cleanup.
        long currentTime = System.currentTimeMillis();
        if (currentTime <= lastBeaconStatusCleanupTime + 1800000L) {
            return;
        }
        for (Iterator<BeaconServerStatus> beaconServerStatusIterator = beaconStatusMap
                .values().iterator(); beaconServerStatusIterator.hasNext();) {
            BeaconServerStatus beaconServerStatus = beaconServerStatusIterator
                    .next();
            // We remove the beacon server status if it has been at least 30
            // minutes since we received the last beacon message and there are
            // no listeners registered for this server.
            if (currentTime - beaconServerStatus.getLastTime() >= 1800000L
                    && !beaconServerStatus.hasBeaconAnomalyListeners()) {
                beaconServerStatusIterator.remove();
            }
        }
        lastBeaconStatusCleanupTime = currentTime;
    }

    private void updateAnomalyListeners() {
        Pair<BeaconAnomalyListener, Boolean> registrationRequest;
        while ((registrationRequest = anomalyListenerUpdates.poll()) != null) {
            if (registrationRequest.getRight() == Boolean.TRUE) {
                anomalyListeners.add(registrationRequest.getLeft());
            } else {
                anomalyListeners.remove(registrationRequest.getLeft());
            }
        }
    }

    private void updateBeaconListeners() {
        Triple<InetSocketAddress, BeaconListener, Boolean> registrationRequest;
        while ((registrationRequest = beaconListenerUpdates.poll()) != null) {
            BeaconServerStatus beaconStatus = beaconServerStatusForAddress(
                    registrationRequest.getLeft());
            if (registrationRequest.getRight() == Boolean.TRUE) {
                beaconStatus.addBeaconAnomalyListener(
                        registrationRequest.getMiddle());
            } else {
                beaconStatus.removeBeaconAnomalyListener(
                        registrationRequest.getMiddle());
            }
        }
    }

    private BeaconServerStatus beaconServerStatusForAddress(
            InetSocketAddress serverAddress) {
        BeaconServerStatus beaconStatus = beaconStatusMap.get(serverAddress);
        if (beaconStatus == null) {
            beaconStatus = new BeaconServerStatus(beaconDetectorStartTime,
                    configuration.getErrorHandler(), serverAddress);
            beaconStatusMap.put(serverAddress, beaconStatus);
        }
        return beaconStatus;
    }

    private void registerWithRepeater() {
        try {
            // We cannot know whether the packet was sent: When in non-blocking
            // mode, the send method returns zero to indicate that the packet
            // was not sent, but it also returns zero when it sent a packet of
            // size zero. We have to wait for a confirmation message to find
            // out.
            // We only try to connect to the loopback address.
            // The C library alternates between the loopback address and the
            // address of the first non-loopback inteface. However, this is only
            // for supporting very old EPICS releases (3.13 beta 11 and older).
            // These releases are very unlikely to run on a computer that also
            // runs this library, so we do not support it any longer.
            channel.send(EMPTY_BYTE_BUFFER, repeaterAddress);
        } catch (IOException e) {
            configuration.getErrorHandler().handleError(
                    DefaultBeaconDetector.class, e,
                    "Could not send registration mesage to repeater.");
        }
    }

    private void startRepeater() {
        if (repeaterRunning()) {
            return;
        }
        // If we cannot start the external repeater we test a second time before
        // we try to start an internal repeater because there is a race
        // condition when starting the repeater.
        if (!startExternalRepeater() && !repeaterRunning()) {
            // We could not start the external repeater and there is still no
            // repeater running, so we start a repeater within this Java VM.
            new Repeater(configuration.getErrorHandler(),
                    configuration.getRepeaterPort()).start();
        }
    }

    private boolean startExternalRepeater() {
        ProcessBuilder pb = new ProcessBuilder(findRepeaterExecutable());
        // Ensure that the repeater is listening on the right port.
        pb.environment().put("EPICS_CA_REPEATER_PORT", Integer.toString(5065));
        try {
            Process p = pb.start();
            // We try to close all streams. The repeater does not perform any
            // I/O through these file descriptors, so keeping them open is just
            // a waste of resources.
            try {
                p.getInputStream().close();
                p.getErrorStream().close();
                p.getOutputStream().close();
            } catch (IOException e) {
                // We ignore any exception that is thrown while trying to close
                // the streams.
            }
            try {
                p.exitValue();
                return false;
            } catch (IllegalThreadStateException e) {
                // We actually expect the exception because the repeater should
                // be running.
                return true;
            }
        } catch (IOException e) {
            // This exception is most likely caused because the caRepeater
            // executable does not exist.
            return false;
        }
    }

    private static String findRepeaterExecutable() {
        String pathEnv = System.getenv("PATH");
        if (pathEnv != null) {
            StringTokenizer tokenizer = new StringTokenizer(pathEnv,
                    File.pathSeparator);
            while (tokenizer.hasMoreTokens()) {
                File directory = new File(tokenizer.nextToken());
                File executable = new File(directory, "caRepeater");
                if (executable.canExecute()) {
                    return executable.getAbsolutePath();
                } else if (File.separatorChar == '\\') {
                    // On Windows, the repeater might be named "caRepeater.exe".
                    File windowsExecutable = new File(directory,
                            "caRepeater.exe");
                    if (windowsExecutable.exists()) {
                        return windowsExecutable.getAbsolutePath();
                    }
                }
            }
        }
        String epicsBaseEnv = System.getenv("EPICS_BASE");
        String epicsHostArchEnv = System.getenv("EPICS_HOST_ARCH");
        if (epicsBaseEnv != null && epicsHostArchEnv != null) {
            File binDir = new File(epicsBaseEnv, "bin");
            File hostArchDir = new File(binDir, epicsHostArchEnv);
            File executable = new File(hostArchDir, "caRepeater");
            if (executable.canExecute()) {
                return executable.getAbsolutePath();
            } else if (File.separatorChar == '\\') {
                // On Windows, the repeater might be named "caRepeater.exe".
                File windowsExecutable = new File(hostArchDir,
                        "caRepeater.exe");
                if (windowsExecutable.exists()) {
                    return windowsExecutable.getAbsolutePath();
                }
            }
        }
        return "caRepeater";
    }

    private boolean repeaterRunning() {
        try {

            DatagramChannel channel = DatagramChannel
                    .open(StandardProtocolFamily.INET);
            try {
                channel.bind(
                        new InetSocketAddress(configuration.getRepeaterPort()));
                channel.setOption(StandardSocketOptions.SO_REUSEADDR, true);
            } finally {
                channel.close();
            }
            return false;
        } catch (BindException e) {
            return true;
        } catch (IOException e) {
            // We log other exceptions but treat them as the repeater running.
            // Even if it is not running, we probably cannot start one if we
            // cannot create and bind a socket.
            configuration.getErrorHandler().handleError(
                    DefaultBeaconDetector.class, e,
                    "Unexpected socket exception while creating a test socket for the repeater port.");
            return true;
        }
    }

    private void processIO() {
        // We update the internal list of listeners by processing the add and
        // remove requests that have been queued since the last run.
        updateAnomalyListeners();
        updateBeaconListeners();
        // Occasionally, we want to remove BeaconServerStatus objects for
        // servers that have not sent a beacon message in a long time and that
        // do not have listeners attached.
        cleanUpBeaconStatusMap();
        // We try to read packets until there is no more data.
        try {
            while (channel.receive(receiveBuffer) != null) {
                receiveBuffer.flip();
                processReceivedPacket();
            }
        } catch (IOException e) {
            // We log the exception and do not continue because most likely,
            // the receive operation would just throw another exception.
            configuration.getErrorHandler().handleError(this.getClass(), e,
                    "Error while trying to receive a UDP packet.");
        } finally {
            // Clear the receive buffer so that it is ready for the next
            // time.
            receiveBuffer.clear();
        }
    }

    private void processReceivedPacket() {
        receiveSource.appendBuffer(receiveBuffer);
        while (receiveSource.remaining() != 0) {
            try {
                try {
                    ChannelAccessMessage message = codec
                            .decodeMessageFromRepeater(receiveSource,
                                    ChannelAccessVersion.NEWEST_SUPPORTED_VERSION);
                    processReceivedMessage(message);
                } catch (MalformedMessageException e) {
                    // We received a message that is not aligned correctly.
                    configuration.getErrorHandler().handleError(
                            DefaultBeaconDetector.class, e,
                            "UDP packet from repeater contained a malformed message.");
                    codec.discardNextMessage(receiveSource);
                } catch (UnsupportedMessageTypeException e) {
                    // We received a message of a type that is not expected from
                    // the repeater.
                    configuration.getErrorHandler().handleError(
                            DefaultBeaconDetector.class, e,
                            "UDP packet from repeater contained a message of unexpected type.");
                    codec.discardNextMessage(receiveSource);
                }
            } catch (BufferUnderflowException e) {
                // There is not enough data left in the packet to read a
                // complete message.
                configuration.getErrorHandler().handleError(
                        DefaultBeaconDetector.class, null,
                        "UDP packet from repeater contained an incomplete message.");
            } catch (Throwable e) {
                // There was some other problem
                configuration.getErrorHandler().handleError(
                        DefaultBeaconDetector.class, e,
                        "Unexpected exception while trying to decode UDP packet from repeater.");
            } finally {
                // Drain the receive source so that the byte buffer is released
                // and can be reused.
                receiveSource.skip(receiveSource.remaining());
            }
        }
    }

    private void processReceivedMessage(ChannelAccessMessage message) {
        switch (message.getCommand()) {
        case CA_PROTO_VERSION:
            // The repeater occasionally sends a dummy message to test whether
            // a client is still alive. We can simply ignore such a message.
            return;
        case CA_PROTO_RSRV_IS_UP:
            processBeaconMessage((ChannelAccessBeaconMessage) message);
            // We received a beacon, thus we can increment the timre for the
            // next repeater check. Typically, a Channel Access server sends a
            // beacon message every 15 seconds, so we wait up to 60 seconds.
            latestNextMessageTime = System.currentTimeMillis() + 60000L;
            return;
        case REPEATER_CONFIRM:
            // We registered successfully, thus we can increment the timer for
            // the next repeater check. Typically, a Channel Access server sends
            // a beacon message every 15 seconds, so we wait up to 60 seconds.
            latestNextMessageTime = System.currentTimeMillis() + 60000L;
            return;
        default:
            configuration.getErrorHandler().handleError(UdpResolver.class, null,
                    "Received unexpected message from repeater: "
                            + message.toString());
        }
    }

    private void processBeaconMessage(ChannelAccessBeaconMessage message) {
        int serverPort = message.getServerPort();
        if (serverPort < 1 || serverPort > 65535) {
            serverPort = configuration.getDefaultServerPort();
        }
        InetSocketAddress serverAddress = new InetSocketAddress(
                message.getServerAddress(), serverPort);
        BeaconServerStatus beaconServerStatus = beaconServerStatusForAddress(
                serverAddress);
        if (beaconServerStatus.beaconReceived(
                message.getVersion().includes(ChannelAccessVersion.V4_10),
                message.getSequenceNumber())) {
            notifyAnomalyListeners(serverAddress);
        }
    }

    private void notifyAnomalyListeners(InetSocketAddress serverAddress) {
        for (BeaconAnomalyListener listener : anomalyListeners) {
            try {
                listener.beaconAnomalyDetected(serverAddress);
            } catch (Throwable e) {
                configuration.getErrorHandler().handleError(
                        DefaultBeaconDetector.class, e,
                        "A beacon anomaly listener threw an exception.");
            }
        }
    }

    private void processTimer() {
        // If the beacon detector just got started or we have not received a
        // message in a long time, we want to start the repeater (if it is not
        // already running) and register with it. This could also happen if no
        // server is on the network or we do not receive beacon messages for a
        // different reason, however it does not hurt to check the repeater
        // occasionally.
        if (System.currentTimeMillis() >= latestNextMessageTime) {
            // Check whether the repeater is running and start it if needed.
            startRepeater();
            // Register with the repeater.
            registerWithRepeater();
            // If this is the first time, we also want to set the start time.
            if (latestNextMessageTime == 0L) {
                beaconDetectorStartTime = System.currentTimeMillis();
            }
            // We wait a short amount of time for the registration confirm
            // message. We do not wait too long, because we might not have been
            // able to send the registration packet and in this case we want to
            // send it again. Five seconds seems enough for the repeater to
            // reply.
            latestNextMessageTime = System.currentTimeMillis() + 5000L;
        }
        // Schedule the next check.
        communicationController.registerTimer(timerProcessor,
                latestNextMessageTime);

    }

    @Override
    public void setCommunicationController(
            CommunicationController communicationController) {
        this.communicationController = communicationController;
        try {
            channelSelectionKey = this.communicationController
                    .registerChannel(new ChannelProcessor() {
                        @Override
                        public void processIO() {
                            DefaultBeaconDetector.this.processIO();
                        }
                    }, channel);
        } catch (ClosedChannelException e) {
            // This can only happen if this object has already been destroyed.
            throw new IllegalStateException(e);
        }
        // We are interested in read operations (receive packets).
        channelSelectionKey.interestOps(SelectionKey.OP_READ);
        // We want to start the repeater (if necessary) and register with it as
        // soon as possible. However, we want to do this in the right thread.
        // Therefore, we schedule a timer.
        this.communicationController.registerTimer(timerProcessor, 0L);
    }

    @Override
    public void addBeaconAnomalyListener(BeaconAnomalyListener listener) {
        if (listener == null) {
            throw new NullPointerException();
        }
        anomalyListenerUpdates.add(Pair.of(listener, Boolean.TRUE));
    }

    @Override
    public void removeBeaconAnomalyListener(BeaconAnomalyListener listener) {
        if (listener == null) {
            throw new NullPointerException();
        }
        anomalyListenerUpdates.add(Pair.of(listener, Boolean.FALSE));
    }

    @Override
    public void addBeaconListener(InetSocketAddress serverAddress,
            BeaconListener listener) {
        if (serverAddress == null || listener == null) {
            throw new NullPointerException();
        }
        beaconListenerUpdates
                .add(Triple.of(serverAddress, listener, Boolean.TRUE));
    }

    @Override
    public void removeBeaconListener(InetSocketAddress serverAddress,
            BeaconListener listener) {
        if (serverAddress == null || listener == null) {
            throw new NullPointerException();
        }
        beaconListenerUpdates
                .add(Triple.of(serverAddress, listener, Boolean.FALSE));
    }

    /**
     * Destroys this beacon detector. This closes the I/O channels opened by
     * this beacon detector. The {@link CommunicationController} that is running
     * the communication thread should be stopped before calling this method.
     * Otherwise, unexpected exceptions might be thrown in the communication
     * thread. The beacon detector is not usable after calling this method and
     * should be discarded.
     */
    public void destroy() {
        try {
            channel.close();
        } catch (IOException e) {
            // We ignore any exception that might be thrown while closing the
            // channel.
        }
    }

}
