/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.io;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.nio.BufferUnderflowException;
import java.nio.channels.SocketChannel;
import java.nio.charset.Charset;
import java.util.concurrent.ConcurrentLinkedQueue;

import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.io.internal.ChannelAccessMessageQueue;
import com.aquenos.epics.jackie.client.io.internal.ConnectionEchoWatchdog;
import com.aquenos.epics.jackie.common.exception.BufferUnderflowWithSizeException;
import com.aquenos.epics.jackie.common.exception.ChannelAccessException;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.exception.MalformedMessageException;
import com.aquenos.epics.jackie.common.exception.ReceivedMessageTooLargeException;
import com.aquenos.epics.jackie.common.exception.SerializedPayloadTooLargeException;
import com.aquenos.epics.jackie.common.exception.UnsupportedInThisVersionException;
import com.aquenos.epics.jackie.common.exception.UnsupportedMessageTypeException;
import com.aquenos.epics.jackie.common.io.AbstractSocketChannelConnection;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.TimerProcessor;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEchoMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessHostNameMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessageCodec;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessUserNameMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersionTCPClientMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersionTCPServerMessage;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueCodec;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Base class for a connection with a Channel Access server.
 * </p>
 * 
 * <p>
 * This class takes care of establishing the connection, negotiating the Channel
 * Access protocol version, and monitoring the connection, sending echo requests
 * when needed. If the server does not support Channel Access version 4.4 or
 * newer, this class immediately destroys the connection.
 * </p>
 * 
 * <p>
 * With the exception of the {@link #destroy()}, {@link #getVersion()}, and
 * {@link #sendMessage(ChannelAccessMessage)} methods, this class is not safe
 * for use by multiple threads and its method may only be called within the
 * communication controller's thread.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public abstract class ChannelAccessServerConnection
        extends AbstractSocketChannelConnection {

    private static final int BUFFER_SIZE = 16384;

    private ChannelAccessMessageCodec codec = new ChannelAccessMessageCodec();
    private CommunicationController communicationController;
    private boolean echoRequestInQueue = false;
    private ConnectionEchoWatchdog echoWatchdog;
    private ErrorHandler errorHandler;
    private final String hostname;
    private ConcurrentLinkedQueue<ChannelAccessMessage> messageProcessQueue = new ConcurrentLinkedQueue<ChannelAccessMessage>();
    private TimerProcessor messageProcessQueueTimer = new TimerProcessor() {
        @Override
        public void processTimer() {
            // We check again, whether the connection has been destroyed in the
            // meantime.
            if (isDestroyed()) {
                // Discard all messages. They cannot be sent if the connection
                // has been destroyed.
                messageProcessQueue.clear();
            } else {
                ChannelAccessMessage message;
                while ((message = messageProcessQueue.poll()) != null) {
                    sendMessageInternal(message);
                }
            }
        }
    };
    private ChannelAccessMessageQueue messageSendQueue = new ChannelAccessMessageQueue();
    private InetSocketAddress serverAddress;
    private final String username;
    // We make the version field volatile so that it can be read from all
    // threads and not just the communication thread.
    private volatile ChannelAccessVersion version;

    /**
     * Creates a connection. This creates the socket channel and tries to
     * connect it. However, the constructor does not wait for the connection to
     * be established.
     * 
     * @param beaconDetector
     *            beacon detector for receiving beacons from the server. Those
     *            beacons are used to assess the connection's health. If
     *            <code>null</code>, beacons are not used at all.
     * @param charset
     *            character set to be used for encoding and decoding messages
     *            (never <code>null</code>).
     * @param communicationController
     *            communication controller with which the socket channel is
     *            registered (never <code>null</code>). This is the
     *            communication controller that drives all I/O and timer based
     *            processing of the connection.
     * @param echoIntervalMillis
     *            interval between sending two echo requests (in milliseconds).
     *            An echo request is only sent if no beacon is received within
     *            this interval. Must be greater than or equal to 100.
     * @param errorHandler
     *            error handler for logging unexpected errors (never
     *            <code>null</code>).
     * @param maxReceiveSize
     *            maximum size of a received Channel Access message in bytes
     *            (including its headers). Must be at least 16400 bytes. When a
     *            larger message is received, the connection is destroyed.
     * @param maxSendSize
     *            maximum size of a sent Channel Access message in bytes
     *            (including its headers). Must be at least 16400 bytes.
     * @param serverAddress
     *            IP address and port number of the server to connect to (never
     *            <code>null</code>).
     * @param hostName
     *            hostname to be sent to the server for authentication. If this
     *            parameter is null, neither the username nor the hostname are
     *            sent.
     * @param userName
     *            username to be sent to the server for authentication. If this
     *            parameter is null, neither the username nor the hostname are
     *            sent.
     * @param version
     *            version of the Channel Access protocol to assume until the
     *            version is received from the server (never <code>null</code>).
     *            Must be at least Channel Access version 4.4. When the server
     *            sends a version that is older than Channel Access version 4.4,
     *            the connection is destroyed immediately.
     * @throws IOException
     *             if the channel for the connection cannot be created.
     * @throws IllegalArgumentException
     *             if the specified <code>version</code> is too old,
     *             <code>maxReceiveSize</code> or <code>maxSendSize</code> are
     *             too small, or <code>echoIntervalMillis</code> is too small.
     * @since EPICS Jackie 2.0.0
     */
    public ChannelAccessServerConnection(BeaconDetector beaconDetector,
            Charset charset, CommunicationController communicationController,
            long echoIntervalMillis, ErrorHandler errorHandler,
            int maxReceiveSize, int maxSendSize,
            InetSocketAddress serverAddress, String hostName, String userName,
            ChannelAccessVersion version) throws IOException {
        super(createChannel(serverAddress), communicationController,
                errorHandler, BUFFER_SIZE,
                (Math.max(maxReceiveSize, maxSendSize) / BUFFER_SIZE) + 1,
                maxReceiveSize * 10L);
        if (beaconDetector == null || charset == null
                || communicationController == null || errorHandler == null
                || serverAddress == null || version == null) {
            throw new NullPointerException();
        }
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            throw new IllegalArgumentException("Channel Access version "
                    + version + " is not supported by this library.");
        }
        if (maxReceiveSize < 16384 + ChannelAccessConstants.STANDARD_HEADER_SIZE
                || maxSendSize < 16384
                        + ChannelAccessConstants.STANDARD_HEADER_SIZE) {
            throw new IllegalArgumentException(
                    "Maximum send and receive sizes must be at least " + (16384
                            + ChannelAccessConstants.STANDARD_HEADER_SIZE)
                            + " bytes.");
        }
        if (echoIntervalMillis < 100L) {
            throw new IllegalArgumentException(
                    "Echo interval must be at least 100 ms.");
        }
        this.codec.setCharset(charset);
        if (maxReceiveSize < 65536
                + ChannelAccessConstants.STANDARD_HEADER_SIZE) {
            codec.setMaxPayloadReceiveSize(maxReceiveSize
                    - ChannelAccessConstants.STANDARD_HEADER_SIZE);
        } else {
            codec.setMaxPayloadReceiveSize(
                    maxReceiveSize - ChannelAccessConstants.LARGE_HEADER_SIZE);
        }
        if (maxSendSize < 65536 + ChannelAccessConstants.STANDARD_HEADER_SIZE) {
            codec.setMaxPayloadSendSize(
                    maxSendSize - ChannelAccessConstants.STANDARD_HEADER_SIZE);
        } else {
            codec.setMaxPayloadSendSize(
                    maxSendSize - ChannelAccessConstants.LARGE_HEADER_SIZE);
        }
        this.communicationController = communicationController;
        this.echoWatchdog = new ConnectionEchoWatchdog(beaconDetector,
                this.communicationController, echoIntervalMillis,
                new ConnectionEchoWatchdog.EchoMessageSender() {
                    @Override
                    public void sendEchoMessage() {
                        ChannelAccessServerConnection.this.sendEchoRequest();
                    }
                }, new ConnectionEchoWatchdog.WatchdogListener() {
                    @Override
                    public void connectionStateChanged(boolean responsive) {
                        if (responsive) {
                            ChannelAccessServerConnection.this
                                    .onConnectionResponsive();
                        } else {
                            ChannelAccessServerConnection.this
                                    .onConnectionUnresponsive();
                        }
                    }
                }, serverAddress);
        this.errorHandler = errorHandler;
        this.hostname = hostName;
        this.serverAddress = serverAddress;
        this.username = userName;
        this.version = version;
    }

    private static SocketChannel createChannel(InetSocketAddress serverAddress)
            throws IOException {
        if (!(serverAddress.getAddress() instanceof Inet4Address)) {
            throw new IllegalArgumentException(
                    "Only IPv4 servers are supported.");
        }
        SocketChannel channel = SocketChannel.open();
        channel.configureBlocking(false);
        Socket socket = channel.socket();
        // The C library uses the SO_TCP_NODELAY and SO_KEEPALIVE options which
        // is very reasonable for this kind of use, so we do the same.
        socket.setTcpNoDelay(true);
        socket.setKeepAlive(true);
        channel.connect(serverAddress);
        return channel;
    }

    private void sendEchoRequest() {
        sendMessage(new ChannelAccessEchoMessage());
        echoRequestInQueue = true;
    }

    @Override
    protected void onConnect() {
        // If this connection has already been destroyed, we do not do anything.
        if (isDestroyed()) {
            return;
        }
        echoWatchdog.start();
        // TODO Make priority configurable.
        sendMessage(new ChannelAccessVersionTCPClientMessage(
                ChannelAccessVersion.NEWEST_SUPPORTED_VERSION, (short) 0));
        sendUserNameAndHostName();
    }

    /**
     * Called when this connection becomes responsive. The connection becomes
     * responsive when the connection's watchdog has earlier detected that it
     * was unresponsive and has now detected that it is responsive again.
     * Initially, a connection is considered responsive and thus this method is
     * only called after {@link #onConnectionUnresponsive()} has been called
     * before. This method is guaranteed to be called in the communication
     * controller's thread. The implementation in the base class does nothing.
     * Derived classes can override it if they want to perform actions based on
     * the connection's responsiveness.
     */
    protected void onConnectionResponsive() {
        // Can be overridden to trigger an action when the connection becomes
        // responsive.
    }

    /**
     * Called when this connection becomes unresponsive. The connection becomes
     * unresponsive when it has been responsive before and now the watchdog has
     * detected that it is unresponsive. Initially, a connection is always
     * considered responsive, therefore this message is only called if the
     * connection becomes unresponsive after it has been established and before
     * it is destroyed. If the connection becomes responsive again after this
     * method has been called, {@link #onConnectionResponsive()} is called. This
     * method is guaranteed to be called in the communication controller's
     * thread. The implementation in the base class does nothing. Derived
     * classes can override it if they want to perform actions based on the
     * connection's responsiveness.
     */
    protected void onConnectionUnresponsive() {
        // Can be overridden to trigger an action when the connection becomes
        // unresponsive.
    }

    @Override
    protected void onDestroy() {
        // Destroy the echo watchdog. This ensures that no more timer events
        // are scheduled by the watchdog. The watchdog's destroy method can be
        // called from any thread, so we do not have to take any precautions.
        echoWatchdog.destroy();
    }

    @Override
    protected long onReceive() {
        ChannelAccessMessage message;
        boolean messageReceived = false;
        long additionalBytesNeeded = 0L;
        do {
            message = null;
            try {
                message = receiveMessage();
            } catch (BufferUnderflowWithSizeException e) {
                // We know that the byte source has accurate information about
                // the number of bytes remaining. Therefore, the bytes that we
                // need in addition to the ones that we have are the bytes that
                // we need in total minus the bytes that we already have.
                additionalBytesNeeded = e.getBytesNeeded()
                        - getReceiveSource().remaining();
            } catch (BufferUnderflowException e) {
                // There is not enough data available but we do not know exactly
                // how much is missing.
                additionalBytesNeeded = 0L;
            }
            if (message != null) {
                messageReceived = true;
                processMessage(message);
            }
        } while (message != null);
        // The implementation in the C library does this inside the loop, but we
        // can do it outside, because we know that there is no other competing
        // thread.
        if (messageReceived) {
            echoWatchdog.messageReceived();
        }
        return additionalBytesNeeded;
    }

    private ChannelAccessMessage receiveMessage() {
        ByteSource byteSource = getReceiveSource();
        // The remaining() method of the byte source provided by the
        // AbstractSocketChannelConnection is guaranteed to return a proper
        // byte count, meaning that there definitely is not more data available
        // than indicated.
        int remaining = byteSource.remaining();
        // If there are less bytes available than needed for a message header,
        // we do not have to do any processing but can simply wait until those
        // bytes become available.
        if (remaining < ChannelAccessConstants.STANDARD_HEADER_SIZE) {
            throw new BufferUnderflowWithSizeException(
                    ChannelAccessConstants.STANDARD_HEADER_SIZE);
        }
        ChannelAccessMessage message;
        try {
            message = codec.decodeMessageToTCPClient(getReceiveSource(),
                    getVersion(), (Inet4Address) serverAddress.getAddress(),
                    serverAddress.getPort());
        } catch (MalformedMessageException e) {
            // We treat a malformed message the same way the C library does: By
            // closing the connection.
            errorHandler.handleError(this.getClass(), e,
                    "Received a malformed message from a TCP server.");
            destroy();
            return null;
        } catch (ReceivedMessageTooLargeException e) {
            // Typically, this should not happen if we take care that we never
            // request more data than we can process. Destroying the connection
            // seems a bit drastic, but we cannot know whether the message was
            // important so if we just discarded it, we might continue with an
            // inconsistent state.
            errorHandler.handleError(this.getClass(), e,
                    "Received a message from a TCP server that was too large.");
            destroy();
            return null;
        } catch (UnsupportedMessageTypeException e) {
            // For this kind of error the same considerations that we have
            // concerning a too large message apply: We cannot know whether the
            // message was important, so we cannot continue.
            errorHandler.handleError(this.getClass(), e,
                    "Received a message of a type that is not known by this implementation from a TCP server.");
            destroy();
            return null;
        }
        return message;
    }

    private void processMessage(ChannelAccessMessage message) {
        // We handle echo responses ourselves and delegate other messages to
        // the derived class.
        switch (message.getCommand()) {
        case CA_PROTO_VERSION:
            processVersionMessage(
                    (ChannelAccessVersionTCPServerMessage) message);
            break;
        case CA_PROTO_ECHO:
            echoWatchdog.echoMessageReceived();
            break;
        default:
            // We simply ignore all other messages and have them handled by the
            // child class.
        }
        onMessageReceived(message);
    }

    private void processVersionMessage(
            ChannelAccessVersionTCPServerMessage message) {
        ChannelAccessVersion version = message.getVersion();
        // Channel Access older than version 4.4 is not supported by this
        // library.
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            destroy();
        }
        // We do not allow protocol downgrades, because this could cause
        // problems with messages that have already been queued.
        if (!version.includes(this.version)) {
            errorHandler.handleError(this.getClass(), null,
                    "Unexpected protocol downgrade from CA version "
                            + this.version + " to CA version " + version + ".");
            destroy();
        }
        this.version = version;
    }

    /**
     * Called when a Channel Access message is received. Even CA_PROTO_VERSION
     * and CA_PROTO_ECHO messages that are processed by the base class are
     * forwarded to this method after they have been processed. This method is
     * guaranteed to be called in the communication controller's thread.
     * 
     * @param message
     *            message that has been received from the server (never
     *            <code>null</code>).
     */
    protected abstract void onMessageReceived(ChannelAccessMessage message);

    @Override
    protected void onReceiveFinished(boolean moreDataAvailable) {
        // If we are receiving a lot of data, we want to notify the echo
        // watchdog. If an echo response is pending, it might not be received in
        // time because other data is sent first. If there is no echo response
        // pending, we want to delay sending the request because the response
        // might get stuck and as long as we are receiving data, the connection
        // seems to be fine anyway.
        if (moreDataAvailable) {
            echoWatchdog.receiveCongestionDetected();
        }
    }

    @Override
    protected void onSend() {
        // After sending some messages, the amount of data in the byte sink
        // should have decreased and we should be able to add some more messages
        // from the queue.
        processMessageSendQueue();
        // If the message queue is empty and there are no more bytes in the byte
        // sink, we know that all pending messages have been sent. This includes
        // any pending echo request. On the other hand, if the message queue or
        // byte sink is not empty and an echo request has been queued, it is
        // possible that this request has not been sent yet. In this case, we
        // want to wait longer for the echo reply. The fact that the we are
        // still able to send data indicates a good connection anyway.
        if (isSendQueueEmpty()) {
            echoRequestInQueue = false;
        } else if (echoRequestInQueue) {
            echoWatchdog.sendCongestionDetected();
        }
    }

    /**
     * Adds a message to the queue of messages to be sent. The message is sent
     * when the connection is ready to accept data and no other message with a
     * higher priority is pending to be sent. This method is thread safe.
     * 
     * @param message
     *            message to be sent to the server (never <code>null</code>).
     * @throws IllegalArgumentException
     *             if this message is of a type that cannot be sent from a
     *             client to a server via TCP.
     * @throws SerializedPayloadTooLargeException
     *             if the message cannot be sent because the payload would be
     *             larger than the configured maximum payload send size or the
     *             specified protocol version is older than Channel Access
     *             version 4.9 and the payload would be larger than 65528 bytes
     *             or the count exceeds 65535 elements.
     * @throws UnsupportedInThisVersionException
     *             if this message is not support in the protocol version
     *             supported by the server.
     * @throws UnsupportedOperationException
     *             if this message has been constructed by deserializing it from
     *             headers only and thus it is incomplete.
     */
    protected void sendMessage(final ChannelAccessMessage message) {
        // We verify the message before queuing it. This ensures that messages
        // that cannot be serialized are not queued.
        codec.verifyMessageToTCPServer(message, version);
        // If this connection has already been destroyed, there is no sense in
        // queuing up more messages. They will not be sent anyway.
        if (isDestroyed()) {
            return;
        }
        // If we are in the communication thread, we can queue the message.
        // Otherwise, we schedule a timer so that the message can be queued from
        // the right thread.
        if (communicationController.inCommunicationThread()) {
            sendMessageInternal(message);
        } else {
            // We add the message to an internal queue. This ensures that
            // messages that are enqueued by the same thread are sent in the
            // correct order.
            messageProcessQueue.add(message);
            communicationController.registerTimer(messageProcessQueueTimer, 0L);
        }
    }

    private void sendMessageInternal(ChannelAccessMessage message) {
        messageSendQueue.add(message);
        processMessageSendQueue();
    }

    private void sendUserNameAndHostName() {
        // The CA_PROTO_CLIENT_NAME and CA_PROTO_HOST_NAME messages are
        // supported since CA version 4.1. We do not support versions before CA
        // version 4.4, but we still do this check in case we want to support
        // them at a later point in time.
        if (!getVersion().includes(ChannelAccessVersion.V4_1)) {
            return;
        }
        if (username != null && hostname != null) {
            sendMessage(new ChannelAccessUserNameMessage(username));
            sendMessage(new ChannelAccessHostNameMessage(hostname));

        }
    }

    private void processMessageSendQueue() {
        // We do not want to serialize too many messages at once because once
        // serialized, we cannot change the order of messages any longer. On the
        // other hand, we do not want to serialize too few messages because this
        // will lead to many small TCP packets being sent which hurts the
        // performance. We serialize messages until we have at least 1024 bytes
        // in the send sink. Typically, this amount should fit into a single IP
        // packet and thus we should not have to wait too long until we can send
        // the next packet. If one of the messages is large, we might exceed
        // this size, but this is okay because have to send the large message at
        // some point.
        ByteSink byteSink = getSendSink();
        while (getSendQueueSizeInBytes() < 1024L
                && !messageSendQueue.isEmpty()) {
            // We know that the encode method will not throw an exception
            // because we verified the message before queuing it. A
            // BufferOverflowException will not be thrown because the byte sink
            // that we use automatically acquires additional buffers.
            ChannelAccessMessage message = messageSendQueue.poll();
            codec.encodeMessageToTCPServer(byteSink, message, version);
        }
    }

    /**
     * Tells whether the send queue is empty. The send queue is empty if the
     * send buffer does not contain unsent data and there are no messages queued
     * to be sent.
     * 
     * @return <code>true</code> if the send queue is empty, <code>false</code>
     *         if there is still data to be sent.
     */
    protected boolean isSendQueueEmpty() {
        return getSendQueueSizeInBytes() == 0L && messageSendQueue.isEmpty();
    }

    /**
     * Returns the Channel Access protocol version supported by the server. If
     * the server has not sent a CA_PROTO_VERSION message yet, this is the
     * version that has been specified when creating this connection. This
     * method is thread safe.
     * 
     * @return Channel Access protocol version supported by the server (never
     *         <code>null</code>).
     */
    protected ChannelAccessVersion getVersion() {
        return version;
    }

    /**
     * Checks whether a value of the specified type and with the specified
     * number of elements can be received without the maximum payload size for
     * received messages.
     * 
     * @param dataType
     *            type of the value that shall be received.
     * @param count
     *            number of elements that shall be received.
     * @throws ChannelAccessException
     *             with a status of
     *             {@link com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus.StatusMessage#ECA_TOLARGE
     *             ECA_TOLARGE} if the payload size would exceed the configured
     *             maximum payload size for received messages.
     */
    protected void verifyResponsePayloadSize(ChannelAccessValueType dataType,
            int count) {
        if (version.includes(ChannelAccessVersion.V4_9)) {
            // Since Channel Access version 4.9, the payload size is only
            // limited by the settings.
            int maxCount = ChannelAccessValueCodec.calculateMaxCount(dataType,
                    codec.getMaxPayloadReceiveSize());
            if (count > maxCount || count < 0) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_TOLARGE
                                .toStatus());
            }
        } else {
            // For older Channel Access versions, the payload size cannot be
            // greater than 65528 and the count cannot be greater than 65535
            // elements.
            int maxPayloadSize = Math.min(codec.getMaxPayloadReceiveSize(),
                    65528);
            int maxCount = ChannelAccessValueCodec.calculateMaxCount(dataType,
                    maxPayloadSize);
            if (count > maxCount || count > 65535 || count < 0) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_TOLARGE
                                .toStatus());
            }
        }
    }
}
