/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver.internal;

/**
 * <p>
 * Estimator for the round-trip time of channel search requests.
 * </p>
 * 
 * <p>
 * This class implements the estimation in the same way the C library does. This
 * algorithm is not perfect, but it has proven to be reliable. In particular, it
 * is on the safe side when it comes to avoiding overloading the network with
 * broadcasts.
 * </p>
 * 
 * <p>
 * This class is only intended for use by {@link DefaultChannelNameResolverImpl}
 * and {@link SearchTimer}. It is closely tied to those classes and has only
 * been separated into a class of its own to keep those classes smaller.
 * </p>
 * 
 * <p>
 * This class is not thread safe, but the instantiating
 * {@link DefaultChannelNameResolverImpl} ensures that it is only used in a safe
 * way.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class RoundTripTimeEstimator {

    /**
     * Minimum estimate for the round-trip time in milliseconds. This is the
     * smallest value that will ever be used as an estimate and - at the same
     * time - the start value.
     */
    private static final double MINIMUM_ESTIMATE = 32.0;

    /**
     * Maximum estimate for the round trip time in milliseconds. The estimator
     * never uses an estimate that is greater than this value. However,
     * {@link #getEstimate()} might still return greater values because the
     * calculated deviation is added to this value.
     */
    private static final double MAXIMUM_ESTIMATE = 30000.0;

    private double estimate = MINIMUM_ESTIMATE;
    private double estimateDeviation = 0.0;

    /**
     * Updates the round-trip-time estimate using the specified measurement.
     * 
     * @param measuredRoundTripTime
     *            measured time between sending a channel search request and the
     *            receiving the response to this request.
     */
    public void updateEstimate(double measuredRoundTripTime) {
        // The measurement is capped to the minimum and maximum values for the
        // estimate. This serves two purposes: First, it guarantees that the
        // estimate can never be outside these limits. Second, it ensures that
        // the impact of a single outlier is not too big.
        if (measuredRoundTripTime > MAXIMUM_ESTIMATE) {
            measuredRoundTripTime = MAXIMUM_ESTIMATE;
        } else if (measuredRoundTripTime < MINIMUM_ESTIMATE) {
            measuredRoundTripTime = MINIMUM_ESTIMATE;
        }
        // The error is the difference between the measured round-trip time and
        // our estimate (how much is the estimation off?).
        double error = measuredRoundTripTime - estimate;
        // We change the estimate by 12.5 percent of the error. This ensures
        // that the estimate does not jump too much, but it still allows for
        // the estimate to eventually reach the measurements.
        estimate += 0.125 * error;
        // The estimate deviation is a measure for how large the difference
        // between our estimation and the measurements is. If the error for this
        // measurement was smaller than the deviation, the deviation is
        // decreased. If it was greater than the deviation, the deviation is
        // increased.
        if (error < 0.0) {
            error = -error;
        }
        estimateDeviation += 0.25 * (error - estimateDeviation);
    }

    /**
     * Returns the current estimate for the round-trip time. The value returned
     * is the sume of the internally calculated estimate (bounded by
     * {@link #MINIMUM_ESTIMATE} and {@link #MAXIMUM_ESTIMATE}) and four times
     * the current estimate deviation. The estimate deviation is calculated
     * based on the difference between the estimate and the actually measured
     * values.
     * 
     * @return estimate for the round-trip time calculated exactly the same way
     *         the Channel Access C library does it.
     */
    public double getEstimate() {
        // The estimate actually used is increased by four time its deviation.
        // This makes sense if the estimate is too small and thus the deviation
        // is large. In this case, it will ensure that the actually used
        // estimate is still large enough. However, if the deviation is large
        // because the estimate is too large, it will actually correct the
        // estimate in the wrong direction. Anyway, this is how it is
        // implemented in the C library. As this implementation works quite
        // well, we still use it, even though it seems illogical.
        return estimate + 4.0 * estimateDeviation;
    }

}
