/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

/**
 * Base class for representations of a <code>CA_PROTO_VERSION</code> message.
 * This message is mainly used for communicating the supported protocol version,
 * however, depending on the context it might carry additional information. As
 * the meaning of this message is different depending on the context (direction
 * and base protocol) there are different child classes for the different use
 * cases.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_VERSION
 */
public abstract class ChannelAccessVersionMessage extends ChannelAccessMessage {

    /**
     * Channel Access version as enum. This is cached on object creation, so
     * that we can avoid the whole process of looking up the right enum each
     * time the property is read. The extra memory overhead seems to be
     * acceptable in comparison to the gain.
     */
    protected ChannelAccessVersion version;

    /**
     * Constructor for creating a <code>CA_PROTO_VERSION</code> message. This
     * constructor is package-private, so that only classes within the same
     * package can be derived from this class.
     * 
     * @param version
     *            Channel Access version supported by the client.
     */
    ChannelAccessVersionMessage(ChannelAccessVersion version) {
        super(ChannelAccessCommand.CA_PROTO_VERSION);
        assert (version != null);
        this.version = version;
    }

    /**
     * Constructor for creating a <code>CA_PROTO_VERSION</code> message. This
     * constructor is package-private, so that only classes within the same
     * package can be derived from this class.
     * 
     * @param messageHeader
     *            the message header that is used to for creating this message.
     *            This header is passed on to the
     *            {@link ChannelAccessMessage#ChannelAccessMessage(ChannelAccessCommand, ChannelAccessMessageHeader)}
     *            constructor.
     * @param version
     *            Channel Access version supported by the client.
     */
    ChannelAccessVersionMessage(ChannelAccessMessageHeader messageHeader,
            ChannelAccessVersion version) {
        super(ChannelAccessCommand.CA_PROTO_VERSION, messageHeader);
        assert (version != null);
        this.version = version;
    }

    /**
     * Returns the Channel Access version supported by the peer sending this
     * message. This information can be used by the receiving peer to decide
     * which features of the protocol it can use.
     * 
     * @return Channel Access version supported by the sending peer.
     */
    public ChannelAccessVersion getVersion() {
        return version;
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessVersionMessage other = (ChannelAccessVersionMessage) obj;
        return new EqualsBuilder().append(this.version, other.version)
                .isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(version).toHashCode();
    }

}
